@@ -9,6 +9,8 @@ use crate::{
 
 use std::{cell::RefCell, marker::PhantomData, rc::Rc};
 use bit_vec::BitVec;
+use byteorder::{ByteOrder, LittleEndian};
+use lzma::reader::LzmaReader;
 
 pub type RCS<C> = Rc<RefCell<C>>;
 
@@ -27,12 +29,13 @@ pub struct Gate<Fr:PrimeField>(
 pub trait CS: Clone {
     type Fr: PrimeField;
     type LC: AbstractLC<Self::Fr>;
+    type GateIterator: Iterator<Item=Gate<Self::Fr>>;
 
     fn num_gates(&self) -> usize;
     fn num_input(&self) -> usize;
     fn num_aux(&self) -> usize;
     fn get_value(&self, index:Index) -> Option<Num<Self::Fr>>;
-    fn get_gate(&self, index:usize) -> &Gate<Self::Fr>;
+    fn get_gate_iterator(&self) -> Self::GateIterator;
 
     // a*b === c
     fn enforce(a: &CNum<Self>, b: &CNum<Self>, c: &CNum<Self>);
@@ -97,24 +100,26 @@ impl<Fr: PrimeField> BuildCS<Fr> {
 pub struct WitnessCS<'a, Fr: PrimeField> {
     pub values_input: Vec<Num<Fr>>,
     pub values_aux: Vec<Num<Fr>>,
-    pub gates: &'a Vec<Gate<Fr>>,
+    pub num_gates: usize,
+    pub gates_data: &'a[u8],
     pub const_tracker: &'a BitVec,
     pub const_tracker_index: usize
 }
 
 impl<'a, Fr: PrimeField> WitnessCS<'a, Fr> {
-    pub fn new(gates: &'a Vec<Gate<Fr>>, const_tracker: &'a BitVec) -> Self {
+    pub fn new(num_gates:usize, gates_data: &'a[u8], const_tracker: &'a BitVec) -> Self {
         Self {
             values_input: vec![Num::ONE],
             values_aux: vec![],
-            gates,
+            num_gates,
+            gates_data,
             const_tracker,
             const_tracker_index: 0
         }
     }
 
-    pub fn rc_new(gates: &'a Vec<Gate<Fr>>, const_tracker: &'a BitVec) -> RCS<Self> {
-        Rc::new(RefCell::new(Self::new(gates, const_tracker)))
+    pub fn rc_new(num_gates:usize, gates_data: &'a[u8], const_tracker: &'a BitVec) -> RCS<Self> {
+        Rc::new(RefCell::new(Self::new(num_gates, gates_data, const_tracker)))
     }
 }
 
@@ -122,6 +127,7 @@ impl<'a, Fr: PrimeField> WitnessCS<'a, Fr> {
 impl<Fr: PrimeField>  CS for DebugCS<Fr> {
     type Fr = Fr;
     type LC = LC<Fr>;
+    type GateIterator = core::iter::Empty<Gate<Self::Fr>>;
 
     fn num_gates(&self) -> usize {
         self.num_gates
@@ -138,8 +144,8 @@ impl<Fr: PrimeField>  CS for DebugCS<Fr> {
         None
     }
 
-    fn get_gate(&self, _:usize) -> &Gate<Self::Fr> {
-        std::unimplemented!()
+    fn get_gate_iterator(&self) -> Self::GateIterator {
+        std::unimplemented!();
     }
 
     // a*b === c
@@ -164,7 +170,7 @@ impl<Fr: PrimeField>  CS for DebugCS<Fr> {
 
     fn alloc(cs: &RCS<Self>, value: Option<&Num<Self::Fr>>) -> CNum<Self> {
         let mut rcs = cs.borrow_mut();
-        let v = rcs.num_aux;
+        let v = rcs.num_aux as u32;
         rcs.num_aux+=1;
         CNum {
             value: value.cloned(),
@@ -174,12 +180,56 @@ impl<Fr: PrimeField>  CS for DebugCS<Fr> {
     }
 
 }
+
+
+pub struct GateStreamedIterator<Fr:PrimeField, R:std::io::Read>(R,PhantomData<Fr>);
+
+fn read_u32<R:std::io::Read>(r: &mut R) -> std::io::Result<u32> {
+    let mut b = [0; 4];
+    r.read_exact(&mut b)?;
+    Ok(LittleEndian::read_u32(&b))
+}
+
+
+fn read_gate_part<Fr:PrimeField, R:std::io::Read>(r: &mut R) -> std::io::Result<Vec<(Num<Fr>, Index)>> {
+    let sz = read_u32(r)? as usize;
+
+    let item_size = std::mem::size_of::<Fr>() + std::mem::size_of::<u8>() + std::mem::size_of::<u32>();
+    let mut buf = vec![0; sz*item_size];
+    r.read_exact(&mut buf)?;
+    let mut buf_ref = &buf[..];
+    let mut gate_part = Vec::with_capacity(sz);
+    for _ in 0..sz {
+        let a = Num::<Fr>::deserialize(&mut buf_ref)?;
+        let b1 = u8::deserialize(&mut buf_ref)?;
+        let b2 = u32::deserialize(&mut buf_ref)?;
+        let b = match b1 {
+            0 => Index::Input(b2),
+            1 => Index::Aux(b2),
+            _ => return Err(std::io::Error::new(std::io::ErrorKind::InvalidData, "enum elements overflow"))
+        };
+        gate_part.push((a,b));
+    }
+    Ok(gate_part)
+}
+
+impl<Fr:PrimeField, R:std::io::Read> Iterator for GateStreamedIterator<Fr, R> {
+    type Item = Gate<Fr>;
+    fn next(&mut self) -> Option<Self::Item> {
+        let a = read_gate_part(&mut self.0).ok()?;
+        let b = read_gate_part(&mut self.0).ok()?;
+        let c = read_gate_part(&mut self.0).ok()?;
+        Some(Gate(a,b,c))
+    }
+}
+
 impl<'a, Fr: PrimeField> CS for WitnessCS<'a, Fr> {
     type Fr = Fr;
     type LC = ZeroLC;
+    type GateIterator = GateStreamedIterator<Fr, LzmaReader<&'a [u8]>>;
 
     fn num_gates(&self) -> usize {
-        self.gates.len()
+        self.num_gates
     }
 
     fn num_input(&self) -> usize {
@@ -191,13 +241,13 @@ impl<'a, Fr: PrimeField> CS for WitnessCS<'a, Fr> {
 
     fn get_value(&self, index:Index) -> Option<Num<Fr>> {
         match index {
-            Index::Input(i) => Some(self.values_input[i]),
-            Index::Aux(i) => Some(self.values_aux[i])
+            Index::Input(i) => Some(self.values_input[i as usize]),
+            Index::Aux(i) => Some(self.values_aux[i as usize]),
         }
     }
 
-    fn get_gate(&self, index:usize) -> &Gate<Fr> {
-        &self.gates[index]
+    fn get_gate_iterator(&self) -> Self::GateIterator {
+        GateStreamedIterator(LzmaReader::new_decompressor(self.gates_data).unwrap(), PhantomData)
     }
 
     fn enforce(_: &CNum<Self>, _: &CNum<Self>, _: &CNum<Self>) {
@@ -230,6 +280,7 @@ impl<'a, Fr: PrimeField> CS for WitnessCS<'a, Fr> {
 impl<Fr: PrimeField> CS for BuildCS<Fr> {
     type Fr = Fr;
     type LC = LC<Fr>;
+    type GateIterator = std::vec::IntoIter<Gate<Self::Fr>>;
 
     fn num_gates(&self) -> usize {
         self.gates.len()
@@ -246,8 +297,8 @@ impl<Fr: PrimeField> CS for BuildCS<Fr> {
         None
     }
 
-    fn get_gate(&self, index:usize) -> &Gate<Fr> {
-        &self.gates[index]
+    fn get_gate_iterator(&self) -> Self::GateIterator {
+        self.gates.clone().into_iter()
     }
 
     // a*b === c
@@ -258,7 +309,7 @@ impl<Fr: PrimeField> CS for BuildCS<Fr> {
 
     fn inputize(n: &CNum<Self>) {
         let mut rcs = n.get_cs().borrow_mut();
-        let v = rcs.num_input;
+        let v = rcs.num_input as u32;
         rcs.num_input+=1;
         rcs.gates.push(Gate(
             n.lc.to_vec(),
@@ -269,7 +320,7 @@ impl<Fr: PrimeField> CS for BuildCS<Fr> {
 
     fn alloc(cs: &RCS<Self>, _: Option<&Num<Self::Fr>>) -> CNum<Self> {
         let mut rcs = cs.borrow_mut();
-        let v = rcs.num_aux;
+        let v = rcs.num_aux as u32;
         rcs.num_aux+=1;
         CNum {
             value: None,
